import asyncio
import dataclasses
import datetime
import hashlib
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from functools import partial
from http import HTTPStatus
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Type,
    TypedDict,
)

import orjson
from jsondaora import as_typed_dict, jsondaora, typed_dict_asjson

from ..asgi.router import Controller
from ..exceptions import BadRequestError, InvalidTasksRepositoryError
from ..method import MethodType
from ..middlewares import Middlewares
from ..responses import Response, json
from ..route.factory import make_route


logger = logging.getLogger(__name__)


try:
    import aioredis
except Exception:
    aioredis = None


class TaskStatusType(Enum):
    RUNNING = 'running'
    FINISHED = 'finished'
    ERROR = 'error'


@jsondaora
class TaskInfo(TypedDict):
    task_id: str
    start_time: str
    status: str
    signature: str
    args_signature: Optional[str]


@dataclasses.dataclass
class BackgroundTask:
    create: Controller
    get_results: Controller


def make_background_task(
    controller: Callable[..., Any],
    path_pattern: str,
    max_workers: int = 10,
    tasks_repository_uri: Optional[str] = None,
    lock: bool = False,
    lock_args: bool = False,
    middlewares: Optional[Middlewares] = None,
    options: bool = False,
) -> BackgroundTask:
    if asyncio.iscoroutinefunction(controller):
        logger.warning(
            'Async tasks can potentially block your application, use with care. '
            'It use is recommended just for small tasks or non-blocking operations.'
        )

    annotations = getattr(controller, '__annotations__', {})
    result_annotation = annotations.get('return', str)
    signature = hashlib.md5(
        '-'.join(
            [controller.__name__]
            + [
                '-'.join((name, type_.__name__))
                for name, type_ in annotations.items()
            ]
        ).encode()
    ).hexdigest()[:12]
    server_id = uuid.uuid4()
    tasks_repository_builder = get_tasks_repository_builder(
        tasks_repository_uri, server_id, signature
    )

    @jsondaora
    class FinishedTaskInfo(TaskInfo):
        end_time: str
        result: result_annotation  # type: ignore

    create_task = make_create_task(
        controller,
        tasks_repository_builder,
        FinishedTaskInfo,
        max_workers,
        lock,
        lock_args,
        signature,
    )
    get_task_results = make_get_task_results(
        tasks_repository_builder, FinishedTaskInfo
    )

    create_task.__annotations__ = {
        name: type_ for name, type_ in annotations.items() if name != 'return'
    }

    return BackgroundTask(
        make_route(
            path_pattern,
            MethodType.POST,
            create_task,
            route_middlewares=middlewares,
            options=options,
        ).controller,
        make_route(
            path_pattern,
            MethodType.GET,
            get_task_results,
            route_middlewares=middlewares,
            options=options,
        ).controller,
    )


def get_iso_time() -> str:
    return (
        datetime.datetime.now()
        .replace(microsecond=0, tzinfo=datetime.timezone.utc)
        .isoformat()
    )


def get_tasks_repository_builder(
    uri: Optional[str], server_id: uuid.UUID, signature: str
) -> Any:
    if uri is None:

        async def builder() -> SimpleTasksRepository:
            return SimpleTasksRepository(server_id, signature, TASKS_DB)

        return builder

    elif isinstance(uri, str) and uri.startswith('redis://'):
        if aioredis is None:
            raise InvalidTasksRepositoryError("'aioredis' package not found!")

        return partial(
            get_redis_tasks_repository,
            server_id=server_id,
            signature=signature,
            uri=uri,
        )

    raise InvalidTasksRepositoryError(uri)


def make_create_task(
    controller: Callable[..., Any],
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    finished_task_info_cls: Any,
    max_workers: int,
    lock: bool,
    lock_args: bool,
    signature: str,
) -> Callable[..., Coroutine[Any, Any, Response]]:
    executor = ThreadPoolExecutor(max_workers)

    async def create_task(*args: Any, **kwargs: Any) -> Response:
        task_id = str(uuid.uuid4())
        args_signature = None

        if lock_args:
            args_signature = hashlib.md5(
                '-'.join(
                    [str(arg) for arg in args]
                    + [
                        '-'.join((key, str(value)))
                        for key, value in kwargs.items()
                    ]
                ).encode()
            ).hexdigest()[:12]

        tasks_repository = await tasks_repository_builder()
        lock_error = await get_lock_error(
            tasks_repository, lock, signature, args_signature
        )

        if lock_error:
            await tasks_repository.close()
            raise lock_error

        if lock:
            await tasks_repository.lock_signature()
        elif args_signature:
            await tasks_repository.lock_args_signature(args_signature)

        if asyncio.iscoroutinefunction(controller):
            wrapper = make_controller_wrapper(
                tasks_repository_builder,
                task_id,
                lock,
                args_signature,
                finished_task_info_cls,
                controller,
                *args,
                **kwargs,
            )
            task_ = asyncio.create_task(wrapper())
            task_.add_done_callback(lambda f: f.result())

        else:
            done_callback = make_done_callback(
                tasks_repository_builder,
                task_id,
                lock,
                args_signature,
                finished_task_info_cls,
            )
            future = executor.submit(controller, *args, **kwargs)
            future.add_done_callback(done_callback)

        start_time = get_iso_time()

        task = TaskInfo(
            task_id=str(task_id),
            start_time=start_time,
            status=TaskStatusType.RUNNING.value,
            signature=signature,
            args_signature=args_signature,
        )
        await tasks_repository.set(task, task_id, finished_task_info_cls)
        await tasks_repository.close()

        return json(task, status=HTTPStatus.CREATED)

    return create_task


def make_controller_wrapper(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    task_id: str,
    lock: bool,
    args_signature: Optional[str],
    finished_task_info_cls: Any,
    controller: Any,
    *args: Any,
    **kwargs: Any,
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def wrapper() -> Any:
        tasks_repository = await tasks_repository_builder()

        try:
            result = await controller(*args, **kwargs)
            status = TaskStatusType.FINISHED.value
        except Exception as error:
            logger.exception(
                f'server-id={tasks_repository.server_id}; '
                f'signature={tasks_repository.signature}; '
                f'args-signature={args_signature};'
                f'task-id={task_id};'
            )
            result = {
                'error': {
                    'name': type(error).__name__,
                    'args': [str(a) for a in error.args],
                }
            }
            status = TaskStatusType.ERROR.value

        task = await tasks_repository.get(task_id, finished_task_info_cls)
        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=status,
            task_id=task['task_id'],
            start_time=task['start_time'],
            signature=task['signature'],
            args_signature=task['args_signature'],
        )
        await tasks_repository.set(
            finished_task, task_id, finished_task_info_cls
        )

        if lock:
            await tasks_repository.unlock_signature()
        elif args_signature:
            await tasks_repository.unlock_args_signature(args_signature)

        await tasks_repository.close()

    return wrapper


def make_done_callback(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    task_id: str,
    lock: bool,
    args_signature: Optional[str],
    finished_task_info_cls: Any,
) -> Callable[[Any], None]:
    def done_callback(future: Any) -> None:
        policy = asyncio.get_event_loop_policy()
        loop = policy.new_event_loop()
        tasks_repository = loop.run_until_complete(tasks_repository_builder())

        try:
            result = future.result()
            status = TaskStatusType.FINISHED.value
        except Exception:
            logger.exception(
                f'server-id={tasks_repository.server_id}; '
                f'signature={tasks_repository.signature}; '
                f'task-key={task_id};'
            )
            result = None
            status = TaskStatusType.ERROR.value

        task = loop.run_until_complete(
            tasks_repository.get(task_id, finished_task_info_cls)
        )

        finished_task = finished_task_info_cls(
            end_time=get_iso_time(),
            result=result,
            status=status,
            task_id=task['task_id'],
            start_time=task['start_time'],
            signature=task['signature'],
        )

        loop.run_until_complete(
            tasks_repository.set(
                finished_task, task_id, finished_task_info_cls,
            )
        )

        if lock:
            loop.run_until_complete(tasks_repository.unlock_signature())
        elif args_signature:
            loop.run_until_complete(
                tasks_repository.unlock_args_signature(args_signature)
            )

        loop.run_until_complete(tasks_repository.close())

    return done_callback


def make_get_task_results(
    tasks_repository_builder: Callable[[], Awaitable['BaseTasksRepository']],
    finished_task_info_cls: Any,
) -> Callable[..., Coroutine[Any, Any, TaskInfo]]:
    async def get_task_results(task_id: str) -> finished_task_info_cls:  # type: ignore
        tasks_repository = await tasks_repository_builder()

        try:
            task = await tasks_repository.get(task_id, finished_task_info_cls)
            return task  # type: ignore
        except KeyError:
            raise BadRequestError(
                name='invalid-task-id', info={'task_id': task_id}
            )
        except ValueError as error:
            if error.args == ('badly formed hexadecimal UUID string',):
                raise BadRequestError(
                    name='invalid-task-id', info={'task_id': task_id}
                )

            raise error from None
        finally:
            await tasks_repository.close()

    return get_task_results


async def get_lock_error(
    tasks_repository: 'BaseTasksRepository',
    lock: bool,
    signature: str,
    args_signature: Optional[str],
) -> Optional[BadRequestError]:
    if lock and await tasks_repository.is_signature_locked():
        return BadRequestError('lock', {'signature': signature})

    if args_signature and await tasks_repository.is_args_locked(
        args_signature
    ):
        return BadRequestError('lock-args', {'args_signature': args_signature})

    return None


@dataclasses.dataclass
class BaseTasksRepository:
    server_id: uuid.UUID
    signature: str
    data_source: Any

    async def set(
        self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
    ) -> None:
        raise NotImplementedError()

    async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
        raise NotImplementedError()

    async def lock_signature(self) -> None:
        raise NotImplementedError()

    async def lock_args_signature(self, args_signature: str) -> None:
        raise NotImplementedError()

    async def is_signature_locked(self) -> bool:
        ...

    async def is_args_locked(self, args_signature: str) -> bool:
        ...

    async def unlock_signature(self) -> Any:
        raise NotImplementedError()

    async def unlock_args_signature(self, args_signature: str) -> Any:
        raise NotImplementedError()

    async def close(self) -> None:
        ...

    def build_signature_key(self) -> str:
        return f'apidaora:{self.server_id}:{self.signature}'

    def build_key(self, task_id: str) -> str:
        return f'{self.build_signature_key()}:{task_id}'


@dataclasses.dataclass
class SimpleTasksRepository(BaseTasksRepository):
    data_source: Dict[str, Any]

    async def set(
        self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
    ) -> None:
        self.data_source[self.build_key(task_id)] = value

    async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
        return self.data_source[self.build_key(task_id)]

    async def lock_signature(self) -> None:
        self.data_source[self.build_signature_key()] = True

    async def lock_args_signature(self, args_signature: str) -> None:
        self.data_source[self.build_key(args_signature)] = True

    async def is_signature_locked(self) -> bool:
        return bool(self.data_source.get(self.build_signature_key(), False))

    async def is_args_locked(self, args_signature: str) -> bool:
        return bool(
            self.data_source.get(self.build_key(args_signature), False)
        )

    async def unlock_signature(self) -> Any:
        return self.data_source.pop(self.build_signature_key())

    async def unlock_args_signature(self, args_signature: str) -> Any:
        return self.data_source.pop(self.build_key(args_signature))


if aioredis is not None:

    @dataclasses.dataclass
    class RedisTasksRepository(BaseTasksRepository):
        data_source: aioredis.Redis

        async def set(
            self, value: Any, task_id: str, task_cls: Type[Any] = TaskInfo,
        ) -> None:
            await self.data_source.set(
                self.build_key(task_id), typed_dict_asjson(value, task_cls)
            )

        async def get(self, task_id: str, finished_task_cls: Type[Any]) -> Any:
            value = await self.data_source.get(self.build_key(task_id))

            if value:
                value = orjson.loads(value)

                if value['status'] == TaskStatusType.RUNNING.value:
                    return as_typed_dict(value, TaskInfo)

                if value['status'] == TaskStatusType.FINISHED.value or (
                    value['status'] == TaskStatusType.ERROR.value
                ):
                    return as_typed_dict(value, finished_task_cls)

            raise KeyError(self.build_key(task_id))

        async def lock_signature(self) -> None:
            await self.data_source.set(self.build_signature_key(), 1)

        async def lock_args_signature(self, args_signature: str) -> None:
            await self.data_source.set(self.build_key(args_signature), 1)

        async def is_signature_locked(self) -> bool:
            return bool(
                await self.data_source.exists(self.build_signature_key())
            )

        async def is_args_locked(self, args_signature: str) -> bool:
            return bool(
                await self.data_source.exists(self.build_key(args_signature))
            )

        async def unlock_signature(self) -> Any:
            return await self.data_source.delete(self.build_signature_key())

        async def unlock_args_signature(self, args_signature: str) -> Any:
            return await self.data_source.delete(
                self.build_key(args_signature)
            )

        async def close(self) -> None:
            self.data_source.close()
            await self.data_source.wait_closed()

    async def get_redis_tasks_repository(
        server_id: uuid.UUID, signature: str, uri: str
    ) -> RedisTasksRepository:
        data_source = await aioredis.create_redis_pool(uri)
        return RedisTasksRepository(server_id, signature, data_source)


TASKS_DB: Dict[str, Any] = {}



import asyncio
from http import HTTPStatus
from logging import getLogger
from typing import Any, Awaitable, Callable, Dict
from urllib import parse

import orjson

from ..exceptions import (
    BadRequestError,
    MethodNotFoundError,
    PathNotFoundError,
)
from .base import ASGIApp, Receiver, Scope, Sender
from .responses import (
    make_json_response,
    send_method_not_allowed_response,
    send_not_found,
    send_response,
)
from .router import ResolvedRoute


logger = getLogger(__name__)


def asgi_app(router: Callable[[str, str], ResolvedRoute]) -> ASGIApp:
    async def controller(
        scope: Scope, receive: Receiver, send: Sender
    ) -> None:
        try:
            resolved = router(scope['path'], scope['method'])

        except PathNotFoundError:
            await send_not_found(send)

        except MethodNotFoundError:
            await send_method_not_allowed_response(send)

        else:
            route = resolved.route

            if route.has_headers:
                headers = scope['headers']
            else:
                headers = []

            try:
                if (
                    hasattr(route.controller, 'middlewares')
                    and route.controller.middlewares
                ):
                    for (
                        middleware
                    ) in route.controller.middlewares.post_routing:
                        middleware(route.path_pattern, resolved.path_args)

            except BadRequestError as error:
                body = orjson.dumps(error.dict)
                await send_response(
                    send,
                    response=make_json_response(
                        content_length=len(body),
                        status=HTTPStatus.BAD_REQUEST,
                        headers=headers,
                    ),
                    body=body,
                )
                return

            if route.has_query:
                query_dict = _get_query_dict(scope)
            else:
                query_dict = {}

            if route.has_body:
                body = await _read_body(receive)
            else:
                body = b''

            response_and_body = route.controller(
                resolved.path_args, query_dict, headers, body
            )

            while asyncio.iscoroutine(response_and_body):
                response_and_body = await response_and_body

            response, body = (
                (response_and_body[0], response_and_body[1])
                if len(response_and_body) > 1
                else (response_and_body[0], b'')
            )
            await send_response(send, response, body)

    return controller


def _get_query_dict(scope: Dict[str, Any]) -> Dict[str, Any]:
    qs = parse.parse_qs(scope['query_string'].decode())
    return qs


async def _read_body(receive: Callable[[], Awaitable[Dict[str, Any]]]) -> Any:
    body = b''
    more_body = True

    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)

    return body



import asyncio
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field
from logging import Logger, getLogger
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Type,
    Union,
)

from .exceptions import BadRequestError
from .header import Header
from .responses import Response
from .route.controller_input import ControllerInput


@dataclass
class MiddlewareRequest:
    path_pattern: str
    path_args: Optional[Dict[str, Any]] = None
    query_dict: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, Header]] = None
    body: Any = None
    kwargs: Optional[Dict[str, Any]] = None


@dataclass
class Middlewares:
    post_routing: List[Callable[[str, Dict[str, Any]], None]] = field(
        default_factory=list
    )
    pre_execution: List[Callable[[MiddlewareRequest], None]] = field(
        default_factory=list
    )
    post_execution: List[
        Callable[[MiddlewareRequest, Response], None]
    ] = field(default_factory=list)


def make_asgi_input_from_requst(
    request: MiddlewareRequest, input_cls: Type[ControllerInput]
) -> ControllerInput:
    kwargs = {}

    if request.path_args:
        kwargs.update(request.path_args)
    if request.query_dict:
        kwargs.update(request.query_dict)
    if request.headers:
        kwargs.update(request.headers)
    if request.body:
        kwargs['body'] = request.body
    if request.kwargs and (input_cls.__annotations_info__.has_kwargs):
        kwargs.update(request.kwargs)

    return input_cls(**kwargs)


class AllowOriginHeader(
    Header, type=str, http_name='access-control-allow-origin'
):
    ...


class ExposeHeadersHeader(
    Header, type=str, http_name='access-control-expose-headers'
):
    ...


class AllowHeadersHeader(
    Header, type=str, http_name='access-control-allow-headers'
):
    ...


class AllowMethodsHeader(
    Header, type=str, http_name='access-control-allow-methods'
):
    ...


class CorsMiddleware:
    def __init__(
        self,
        *,
        allow_origin: Optional[str] = None,
        expose_headers: Optional[str] = None,
        allow_headers: Optional[str] = None,
        allow_methods: Optional[str] = None,
        servers_all: Optional[str] = '*',
    ):
        if (
            not allow_origin
            and not expose_headers
            and not allow_headers
            and not allow_methods
        ):
            self.headers = [
                AllowOriginHeader(servers_all),
                ExposeHeadersHeader(servers_all),
                AllowHeadersHeader(servers_all),
                AllowMethodsHeader(servers_all),
            ]
        else:
            self.headers = []
            if allow_origin:
                self.headers.append(AllowOriginHeader(allow_origin))
            if expose_headers:
                self.headers.append(ExposeHeadersHeader(expose_headers))
            if allow_headers:
                self.headers.append(AllowHeadersHeader(allow_headers))
            if allow_methods:
                self.headers.append(AllowMethodsHeader(allow_methods))

        self.headers_tuple = tuple(self.headers)

    def __call__(self, request: MiddlewareRequest, response: Response) -> None:
        if isinstance(response.headers, list):
            response.headers.extend(self.headers)

        elif isinstance(response.headers, tuple):
            response.headers = response.headers + self.headers_tuple

        elif response.headers is None:
            response.headers = self.headers_tuple


@dataclass
class LockRequestMiddleware:
    locks: Set[str] = field(default_factory=set)

    def lock(self, path_pattern: str, path_args: Dict[str, Any]) -> None:
        if path_pattern in self.locks:
            raise BadRequestError(
                'lock-request', {'path_pattern': path_pattern}
            )

        self.locks.add(path_pattern)

    def unlock_pre_execution(self, request: MiddlewareRequest) -> None:
        self.locks.remove(request.path_pattern)

    def unlock_post_execution(
        self, request: MiddlewareRequest, response: Response
    ) -> None:
        self.locks.remove(request.path_pattern)


@dataclass(init=False)
class BackgroundTaskMiddleware:
    executor: ThreadPoolExecutor
    logger: Logger

    def __init__(
        self, max_workers: int = 100, logger: Optional[Logger] = None
    ):
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

        if logger is None:
            logger = getLogger(__name__)

        self.logger = logger

    def __call__(self, request: MiddlewareRequest, response: Response) -> None:
        response_tasks: Union[
            Iterable[Callable[[], None]], Callable[[], None]
        ] = []

        if response.kwargs:
            response_tasks = response.kwargs.get(
                'background_tasks', response_tasks
            )

        if response_tasks and not isinstance(response_tasks, Iterable):
            response_tasks = (response_tasks,)

        for task in response_tasks:
            future = self.executor.submit(task)
            future.add_done_callback(self.done_callback)

    def done_callback(self, future: Any) -> None:
        try:
            future.result()
        except Exception:
            self.logger.exception('Background Task Error')


@dataclass(init=False)
class AsyncBackgroundTaskMiddleware:
    logger: Logger = getLogger(__name__)

    def __call__(self, request: MiddlewareRequest, response: Response) -> None:
        response_tasks: Union[
            Iterable[Awaitable[None]],
            Iterable[asyncio.Task[None]],
            Awaitable[None],
        ] = []

        if response.kwargs:
            response_tasks = response.kwargs.get(
                'background_tasks_async', response_tasks
            )

        if not isinstance(response_tasks, Iterable):
            response_tasks = (response_tasks,)

        for task in response_tasks:
            if asyncio.iscoroutinefunction(task):  # type: ignore
                task = task()  # type: ignore

            task = asyncio.create_task(task)
            task.add_done_callback(self.done_callback)

    def done_callback(self, future: Any) -> None:
        try:
            future.result()
        except Exception:
            self.logger.exception('Background Task Error')



from typing import Any, Callable, Union

from ..asgi.router import Controller
from ..controllers.background_task import BackgroundTask, make_background_task
from ..exceptions import InvalidRouteArgumentsError, MethodNotFoundError
from ..method import MethodType
from .factory import make_route


class _RouteDecorator:
    def __getattr__(
        self, attr_name: str
    ) -> Callable[
        ..., Callable[[Callable[..., Any]], Union[Controller, BackgroundTask]]
    ]:
        if attr_name == '__name__':
            return type(self).__name__  # type: ignore

        if attr_name == 'background':
            brackground = True

        else:
            brackground = False
            method = attr_name.upper()

            if method not in MethodType.__members__:
                raise MethodNotFoundError(attr_name)

        def decorator(
            path_pattern: str, **kwargs: Any
        ) -> Callable[[Callable[..., Any]], Union[Controller, BackgroundTask]]:
            keys = tuple(kwargs.keys())
            if len(kwargs) > 0 and not any(
                (
                    'tasks_repository_uri' in keys,
                    'lock' in keys,
                    'lock_args' in keys,
                    'middlewares' in keys,
                    'options' in keys,
                )
            ):
                raise InvalidRouteArgumentsError(kwargs)

            def wrapper(
                controller: Callable[..., Any]
            ) -> Union[Controller, BackgroundTask]:
                middlewares = kwargs.get('middlewares')
                options = kwargs.get('options')

                if brackground:
                    tasks_repository_uri = kwargs.get('tasks_repository_uri')
                    lock = kwargs.get('lock')
                    lock_args = kwargs.get('lock_args')
                    return make_background_task(
                        controller,
                        path_pattern,
                        tasks_repository_uri=tasks_repository_uri,
                        lock=lock,  # type: ignore
                        lock_args=lock_args,  # type: ignore
                        middlewares=middlewares,
                        options=options,  # type: ignore
                    )

                else:
                    route = make_route(
                        path_pattern,
                        MethodType[method],
                        controller,
                        route_middlewares=middlewares,
                        options=options,  # type: ignore
                    )
                    return route.controller

            return wrapper

        return decorator


route = _RouteDecorator()
